/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
MoveSwapMatrix: Efficient representation for moving and swapping
around in IBM3 training.

Franz Josef Och (30/07/99)
--*/

#include "MoveSwapMatrix.h"

//increasing efficiency: avoiding double-counting of MoveSwapMatrix elements
//increasing efficiency: maintaining list of possible improvements (for faster access of best improvement
template<class TRANSPAIR>
MoveSwapMatrix<TRANSPAIR>::MoveSwapMatrix(const TRANSPAIR&_ef, const alignment&_a)
  : alignment(_a), ef(_ef), l(ef.get_l()), m(ef.get_m()), _cmove(l+1, m+1), _cswap(m+1, m+1), changed(l+2, 0), changedCounter(1), modelnr(_ef.modelnr())
{
  for(WordIndex j=1;j<=m;j++)updateJ(j, 0);
}

template<class TRANSPAIR>
void MoveSwapMatrix<TRANSPAIR>::updateJ(WordIndex j, bool useChanged)
{
  for(WordIndex i=0;i<=l;i++)
    if( get_al(j)!=i &&(useChanged==0||changed[i]!=changedCounter) )
      _cmove(i, j)=ef.scoreOfMove((*this), i, j);
  for(WordIndex j2=j+1;j2<=m;j2++)
    if( get_al(j)!=get_al(j2) ){
      _cswap(j, j2)=ef.scoreOfSwap((*this), j, j2);
    }
  for(WordIndex j2=1;j2<j;j2++)
    if( get_al(j)!=get_al(j2) ){
      _cswap(j2, j)=ef.scoreOfSwap((*this), j2, j);
    } 
}
template<class TRANSPAIR>
void MoveSwapMatrix<TRANSPAIR>::updateI(WordIndex i)
{
  for(WordIndex j=1;j<=m;j++)
    if( get_al(j)!=i )
      _cmove(i, j)=ef.scoreOfMove((*this), i, j);
}

template<class TRANSPAIR>
void MoveSwapMatrix<TRANSPAIR>::printWrongs()
{
  for(WordIndex i=0;i<=l;i++)
    {
      for(WordIndex j=1;j<=m;j++)
	if( get_al(j)==i)
	  cout << "A";
	else
	  {
	    LogProb real=_cmove(i, j), wanted=ef.scoreOfMove((*this), i, j);
	    if( fabs(1.0-real/wanted)>1e-3 )
	      cout << 'b';
	    else if(fabs(1.0-real/wanted)>1e-10 )
	      cout << 'e';
	    else if(real!=wanted)
	      cout << 'E';
	    else
	      cout << ' ';
	  }
      cout << endl;
    }
  cout << endl;
  for(WordIndex j=1;j<=m;j++)
    {
      for(WordIndex j1=1;j1<=m;j1++)
	if( j1>j )
	  {
	    if( get_al(j)==get_al(j1) )
	      cout << 'A';
	    else 
	      cout << (_cswap(j, j1)==ef.scoreOfSwap((*this), j, j1));
	  }
	else
	  cout << ' ';
      cout << endl;  
    }
}
template<class TRANSPAIR>
bool MoveSwapMatrix<TRANSPAIR>::isRight()
{
  for(WordIndex i=0;i<=l;i++)
    for(WordIndex j=1;j<=m;j++)
      if( get_al(j)!=i && (!(doubleEqual(_cmove(i, j), ef.scoreOfMove((*this), i, j)))) )
	{
	  cerr << "DIFF: " << i << " " << j << " " << _cmove(i, j) << " " << ef.scoreOfMove((*this), i, j) << endl;
	  return 0;
	}
  for(WordIndex j=1;j<=m;j++)
    for(WordIndex j1=1;j1<=m;j1++)
      if( j1>j&&get_al(j)!=get_al(j1)&&(!doubleEqual(_cswap(j, j1), ef.scoreOfSwap((*this), j, j1))) )
	{
	  cerr << "DIFFERENT: " << j << " " << j1 << " " << _cswap(j, j1) << " " << ef.scoreOfSwap((*this), j, j1) << endl;
	  return 0;
	}
  return 1;
}

template<class TRANSPAIR>
void MoveSwapMatrix<TRANSPAIR>::doMove(WordIndex _i, WordIndex _j)
{
  WordIndex old_i=get_al(_j);
#ifdef VDEBUG
  if( !isRight() )
    {
      cerr << "ERROR: before moving at position " << _j << " from " << old_i << " to " << _i << endl;
      printWrongs();
    }
#endif
  if ( modelnr==4 )
    {
      changedCounter++;
      for(unsigned int k=prev_cept(old_i);k<=next_cept(old_i);++k)changed[k]=changedCounter;
      for(unsigned int k=prev_cept(_i);k<=next_cept(_i);++k)changed[k]=changedCounter;
      set(_j, _i);
      for(unsigned int k=prev_cept(old_i);k<=next_cept(old_i);++k)changed[k]=changedCounter;
      for(unsigned int k=prev_cept(_i);k<=next_cept(_i);++k)changed[k]=changedCounter;
      for(unsigned int i=0;i<=l;i++)
	if(changed[i]==changedCounter)
	  updateI(i);
      for(unsigned int j=1;j<=m;j++)
	if( changed[get_al(j)]==changedCounter )
	  updateJ(j, 1);
    }
  else
    {
      assert(modelnr==3);
      set(_j, _i);
      changedCounter++;
      updateI(old_i);
      changed[old_i]=changedCounter;
      updateI(_i);
      changed[_i]=changedCounter;
      for(WordIndex j=1;j<=m;j++)
	if( get_al(j)==_i || get_al(j)==old_i )
	  updateJ(j, 1);
    }
#ifdef VDEBUG
  if( !isRight() )
    {
      cerr << "ERROR: moving at position " << _j << " from " << old_i << " to " << _i << endl;
      printWrongs();
    }
#endif
}
template<class TRANSPAIR>
void MoveSwapMatrix<TRANSPAIR>::doSwap(WordIndex _j1, WordIndex _j2)
{
  assert( _cswap(_j1, _j2)>1 );
  WordIndex i1=get_al(_j1), i2=get_al(_j2);
  if( modelnr==4 )
    {
      changedCounter++;
      for(unsigned int k=prev_cept(i1);k<=next_cept(i1);++k)changed[k]=changedCounter;
      for(unsigned int k=prev_cept(i2);k<=next_cept(i2);++k)changed[k]=changedCounter;
      set(_j1, i2);
      set(_j2, i1);
      for(unsigned int i=0;i<=l;i++)
	if(changed[i]==changedCounter)
	  updateI(i);
     for(unsigned int j=1;j<=m;j++)
	if( changed[get_al(j)]==changedCounter )
	  updateJ(j, 1);
    }
  else
    {
      assert(modelnr==3);
      set(_j1, i2);
      set(_j2, i1);
      changedCounter++;
      updateI(i1);
      changed[i1]=changedCounter;
      updateI(i2);
      changed[i2]=changedCounter;
      updateJ(_j1, 1);
      updateJ(_j2, 1);
    }
#ifdef VDEBUG
  if( !isRight() )
    printWrongs();
#endif
}


#include "transpair_model3.h"
#include "transpair_model4.h"
template class MoveSwapMatrix<transpair_model3>;
template class MoveSwapMatrix<transpair_model4>;
