/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include "model1.h"
#include "Globals.h"
#include "utility.h"


model1::model1(const char* efname, vcbList& evcblist, vcbList& fvcblist):
  efFilename(efname), Elist(evcblist), Flist(fvcblist), 
  eTotalWCount(Elist.totalVocab()), fTotalWCount(Flist.totalVocab()), 
  noEnglishWords(Elist.size()), noFrenchWords(Flist.size()), tTable(), 
  evlist(Elist.getVocabList()), fvlist(Flist.getVocabList())
{}

void model1::initialize_table_uniformly(sentenceHandler& sHandler1){
  WordIndex i, j;

  cerr << "\nInitializing t Table uniformally from corpus \n";

  sentPair sent ;
  //  sentenceHandler sHandler1(efFilename.c_str());
  sHandler1.rewind();
  while(sHandler1.getNextSentence(sent)){
    Vector<WordIndex>& es = sent.eSent;
    Vector<WordIndex>& fs = sent.fSent;
    if ((sent.sentenceNo % 10000) == 0)
      cerr <<sent.sentenceNo << '\n'; 
    PROB uniform = 1.0/es.size() ;
    for( i=0; i < es.size(); i++)
      for(j=1; j < fs.size(); j++)
	tTable.insert(es[i],fs[j],0,uniform);
  }
  cerr << "\ntTable contains " << tTable.getHash().bucket_count() << 
    " buckets, " << tTable.getHash().size() << "entries.\n";
}


void model1::em_with_tricks(int noIterations, Perplexity& perp, sentenceHandler& sHandler1, 
			    bool seedModel1, Dictionary& dictionary, bool useDict, Perplexity* testPerp, sentenceHandler* testHandler, 
			    Perplexity& trainViterbiPerp, Perplexity* testViterbiPerp)
{
  string tfile, tcfile, number, tfile_actual, alignfile, test_alignfile;
  time_t st, it_st, fn, it_fn;
  int pair_no;
  bool dump_files = false ;
  //  Perplexity perp;
  st = time(NULL);
  sHandler1.rewind();
  cerr << "\n==========================================================\n";
  cerr << "Model1 Training Started at: "<< ctime(&st) << "\n";  
  for(int it = 1; it <= noIterations; it++){
    pair_no = 0 ;
    it_st = time(NULL);
    cerr <<  "\n-----------\nModel1: Iteration " << it << '\n';
    dump_files = (Model1_Dump_Freq != 0) &&  ((it % Model1_Dump_Freq)  == 0) && !NODUMPS ;
    if (dump_files){
      // create files to dump tables, if neccessary (when required and at end
      // print the current t table
      number = "";
      int n = it;
      do{
	//mj changed next line to work with new string library
	number.insert((unsigned int)0, 1, (char)(n % 10 + '0'));
      } while((n /= 10) > 0);
      tfile = Prefix + ".t1." + number ;
      tfile_actual = Prefix + ".actual.t1." + number ;
      tcfile = Prefix + ".tc1." + number  ;
      alignfile = Prefix + ".A1." + number ;
      test_alignfile = Prefix +".tst.A1." + number ;
    }
     em_loop(it, perp, sHandler1, seedModel1, dump_files, alignfile.c_str(), dictionary, useDict, trainViterbiPerp); 
     // collect counts for training corpus
     if (testPerp && testHandler) // calculate test perplexity
       em_loop(it, *testPerp, *testHandler, seedModel1, dump_files, test_alignfile.c_str(), dictionary, useDict, *testViterbiPerp, true); 
     if (dump_files){
       //       tTable.printCountTable(tcfile.c_str(), evlist, fvlist, true);
    }
    cerr << "\nModel1 Iteration " << it << " done. I normalize now...";
    tTable.normalizeTable(Elist, Flist);
    cerr << "\ntTable contains " << tTable.getHash().bucket_count() << 
      " buckets, " << tTable.getHash().size() << " entries.\n";
    cerr << "\nModel1: ("<<it<<") TRAIN CROSS-ENTROPY " << perp.cross_entropy()
	 << " PERPLEXITY " << perp.perplexity() << '\n';
     if (testPerp && testHandler)
       cerr << "\nModel1: ("<<it<<") TEST CROSS-ENTROPY " << (*testPerp).cross_entropy()
	    << " PERPLEXITY " << (*testPerp).perplexity() 
	    << '\n';
     cerr << "\nModel1: ("<<it<<") VITERBI TRAIN CROSS-ENTROPY " << trainViterbiPerp.cross_entropy()
	  << " PERPLEXITY " << trainViterbiPerp.perplexity() << '\n';
     if (testPerp && testHandler)
       cerr << "\nModel1: ("<<it<<") VITERBI TEST CROSS-ENTROPY " << (*testViterbiPerp).cross_entropy()
	    << " PERPLEXITY " << (*testViterbiPerp).perplexity() 
	    << '\n';

    if (dump_files){
      tTable.printProbTable(tfile.c_str());
    }    
    it_fn = time(NULL);
    cerr << "\nModel 1 Iteration: " << it<< " took: " << difftime(it_fn, it_st) << " seconds\n";
  } /* of of  iterstions*/
  fn = time(NULL) ;
  //interval  = difftime(fn, st);
  cerr <<  "\nEntire Model1 Training took: " << difftime(fn, st) << " seconds\n";
    cerr << endl << "tTable contains: " << 
      tTable.getHash().bucket_count() << " buckets, " <<
      tTable.getHash().size() << " entries. \n";
  cerr << "\nModel1 Training Finished at: "<< ctime(&fn) << endl;  
  cerr << "==========================================================\n";
}

void model1::load_table(const char* tname){
  /* This function loads the t table from the given file; use it
     when you want to load results from previous t training
     without doing any new training.
     NAS, 7/11/99
  */
  cerr << "Model1: loading t table \n" ;
  tTable.readProbTable(tname);
}

  

void model1::em_loop(int it, Perplexity& perp, sentenceHandler& sHandler1, bool seedModel1, 
		     bool dump_alignment, const char* alignfile, Dictionary& dict, bool useDict, Perplexity& viterbi_perp, bool test)
{
  WordIndex i, j, l, m ;
  double cross_entropy;
  int pair_no=0 ;
  perp.clear();
  viterbi_perp.clear();
  ofstream of2;
  // for each sentence pair in the corpus
  if (dump_alignment)
    of2.open(alignfile);
  PROB uniform = 1.0/noFrenchWords ;
  sentPair sent ;
  sHandler1.rewind();
  while(sHandler1.getNextSentence(sent)){
    Vector<WordIndex>& es = sent.eSent;
    Vector<WordIndex>& fs = sent.fSent;
    const float so  = sent.noOccurrences;
    if ((sent.sentenceNo % 10000 ) == 0)
	cerr << sent.sentenceNo << '\n';
    l = es.size() - 1;
    m = fs.size() - 1;
    cross_entropy = log(1);
    Vector<WordIndex> viterbi_alignment(fs.size());
    double viterbi_score = 1 ;

    bool eindict[l + 1];
    bool findict[m + 1];
    bool indict[m + 1][l + 1];
    if(it == 1 && useDict){
      for(unsigned int dummy = 0; dummy <= l; dummy++) eindict[dummy] = false;
      for(unsigned int dummy = 0; dummy <= m; dummy++){
	findict[dummy] = false;
	for(unsigned int dummy2 = 0; dummy2 <= l; dummy2++) 
	  indict[dummy][dummy2] = false;
      }
      for(j = 0; j <= m; j++)
	for(i = 0; i <= l; i++)
	  if(dict.indict(fs[j], es[i])){
	    //	    cerr << "\tDictionary pair:  F" << fs[j] << " E" << es[i] << endl;
	    eindict[i] = findict[j] = indict[j][i] = true;
	  }
//       cerr << "     ";
//       for(i = 0; i <= l; i++) fprintf(stderr, "%5d", es[i]);
//       cerr << endl;
//       for(j = 0; j <= m; j++){
// 	fprintf(stderr, "%5d", fs[j]);
// 	for(i = 0; i <= l; i++)
// 	  fprintf(stderr, "%5c", 
// 		  ((indict[j][i] || (!eindict[i] && !findict[j]))? 'X' : '.'));
// 	//	fprintf(stderr, "%5d", !findict[j]);
// 	cerr << endl;
//       }
//       //      cerr << "     ";
//       //      for(i = 0; i <= l; i++) fprintf(stderr, "%5d", !eindict[i]);
//       //      cerr << endl;
    }


    for(j=1; j <= m; j++){
      Vector<LpPair<COUNT,PROB> *> sPtrCache(es.size(),0); // cache pointers to table 
      // entries  that map fs to all possible ei in this sentence.
      PROB denom = 0.0;
      WordIndex best_i = 0 ; // i for which fj is best maped to ei
      PROB word_best_score = 0 ;  // score for the best mapping of fj
      if (it == 1 && !seedModel1){
	denom = uniform  * es.size() ;
	word_best_score = uniform ;
      }
      else 
	for(i=0; i <= l; i++){
	  PROB e(0.0) ;
	  sPtrCache[i] = tTable.getPtr(es[i], fs[j]) ;
	  if (sPtrCache[i] != 0 && (*(sPtrCache[i])).prob > PROB_SMOOTH) 
	    // if valid pointer
	    e = (*(sPtrCache[i])).prob;
	  else e = PROB_SMOOTH ;
	  denom += e  ;
	  if (e > word_best_score){
	    word_best_score = e ;
	    best_i = i ;
	  }
	}
      viterbi_alignment[j] = best_i ;
      viterbi_score *= word_best_score ; /// denom ;
      if (denom == 0){
	if (test)
	  cerr << "WARNING: denom is zero (TEST)\n";
	else 
	  cerr << "WARNING: denom is zero (TRAIN)\n";
      }
      cross_entropy += log(denom) ;
      if (!test){
	if(denom > 0){	  
	  COUNT val = COUNT(so) / (COUNT) double(denom) ;
	  /* this if loop implements a constraint on counting:
	     count(es[i], fs[j]) is implemented if and only if
	     es[i] and fs[j] occur together in the dictionary, 
	     OR
	     es[i] does not occur in the dictionary with any fs[x] and
	     fs[j] does not occur in the dictionary with any es[y]
	  */
	  if(it == 1 && useDict){
	    for(i = 0; i <= l; i++){
	      if(indict[j][i] || (!findict[j] && !eindict[i])){
		PROB e(0.0) ;
		if (it == 1 && !seedModel1)
		  e =  uniform  ;
		else if (sPtrCache[i] != 0 &&  (*(sPtrCache[i])).prob > PROB_SMOOTH) 
		  e = (*(sPtrCache[i])).prob;
		else e = PROB_SMOOTH ;
		
		if (sPtrCache[i] != 0)
		  (*(sPtrCache[i])).count += COUNT(e) * val;
		else 	      
		  tTable.incCount(es[i], fs[j], COUNT(e) * val);
	      } /* end of if */
	    } /* end of for i */
	  } /* end of it == 1 */
	  // Old code:
	  else{
	    for(i=0; i <= l; i++) {	    
	      PROB e(0.0) ;
	      if (it == 1 && !seedModel1)
		e =  uniform  ;
	      else if (sPtrCache[i] != 0 &&  (*(sPtrCache[i])).prob > PROB_SMOOTH) 
		e = (*(sPtrCache[i])).prob;
	      else e = PROB_SMOOTH ;
	      if (sPtrCache[i] != 0) 
		(*(sPtrCache[i])).count += COUNT(e) * val;
	      else 	      
		tTable.incCount(es[i], fs[j], COUNT(e) * val);
	    } /* end of for i */
	  } // end of else
	} // end of if (denom > 0)
      }// if (!test)
    } // end of for (j) ;
    perp.addFactor(cross_entropy-m*log(l+1), so, l, m);
    viterbi_perp.addFactor(log(viterbi_score)-m*log(l+1), so, l, m);
      // the term pow(l+1, m) or (m * log(l+1) is for the uniform alignment 
      // probability 
      if (dump_alignment)
      	printAlignToFile(es, fs, evlist, fvlist, of2, viterbi_alignment, sent.sentenceNo, viterbi_score);
      pair_no++;
  } /* of while */
  cerr << "\nProcessed a total of: "<<pair_no<<" sentence pairs.\n"; 
  sHandler1.rewind();
  perp.record(0, it);
  viterbi_perp.record(0, it);
}
