/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include <strstream>
#include <string>
#include <fstream.h>
#include "defs.h"
#include "utility.h"
#include "Globals.h"

void parseConfigFile (char * fname )
  // This functions reads in the configuration file to set up some run-time
  // parameters. The parameters are global variables that are defined in 
  // main.cc and used all over the place in the program
  // The format of the configuration file can be explained in the following way
  // FORMAT:
  // the character '\n' separates lines ..
  // lines that start with "//" (skipping over white spaces are considered 
  // as comments and will be ignored.
  // Any other line is considered as an attribute setting instruction and it 
  // is divided into haves (separated by a colon ":"). The first half is the
  // attribute value which consists of the concatenation of all non-white space
  // tokens before the colon. These tokens will have spaces eseparating them.
  // The attribute vlue is the first token after the colon (any thing after 
  // it will be ignored ;
  // For example :
  // if the configuration file has the following entry:
  //
  // NO.   ITERATIONS   MODEL 2 :	10
  //
  // then the attribute is "NO. ITERATIONS MODEL 2" , and the attribute value
  // is "10"  (these do not include the quotation marks).

{

  string line, word, attrib, attribval ;
  ifstream Config_File(fname);
  if(!Config_File){
    cerr << "ERROR:  Cannot open configuration file " << fname << "!\n" ;
    exit(1);
  }

  cerr << "The following options are from the config file and will be overwritten by any command line options.\n";
  
  while(getline(Config_File, line)){

    istrstream buffer(line.c_str());
    word = attrib = attribval = "" ;
    buffer >> word  ;
    if (word != "//"){ // if line does not start with "//" (i.e. not a comment)
      attrib = word ;
      while((buffer >> word) && (word != ":")){
	attrib += " " + word ;
      }      
      buffer >> attribval ;
      // This is where (1) the configuration file is defined and
      //               (2) parsing of its attributes occurs.
      
      if(attrib == "SOURCE VOCABULARY FILE"){
	SourceVocabFilename = attribval;
	cerr << "\tSource vocabulary file:  " << SourceVocabFilename << '\n';
      }
      else if(attrib == "TARGET VOCABULARY FILE"){
	TargetVocabFilename = attribval;
	cerr << "\tTarget vocabulary file:  " << TargetVocabFilename << '\n';
      }
      else if(attrib == "CORPUS FILE"){
	CorpusFilename = attribval;
	cerr << "\tCorpus file:  " << CorpusFilename << '\n';
      }
      else if(attrib == "TEST CORPUS FILE"){
	TestCorpusFilename = attribval;
	cerr << "\tCorpus file:  " << TestCorpusFilename << '\n';
      }
      else if(attrib == "MAX SENTENCE LENGTH"){
	unsigned int temp = atoi(attribval.c_str()) + 1;
	if (temp > 1 && temp <= MAX_SENTENCE_LENGTH_ALLOWED){
	  MAX_SENTENCE_LENGTH = temp ;
	  cerr << "\tMAX SENTENCE LENGTH:  " << temp << '\n';
	}
      }
      else if(attrib == "NO. ITERATIONS MODEL 1"){
	Model1_Iterations = atoi(attribval.c_str());
	cerr << "\tNo. Model 1 iterations:  " << Model1_Iterations << '\n';
      }
      else if(attrib == "NO. ITERATIONS MODEL 2"){
	Model2_Iterations = atoi(attribval.c_str());
	cerr << "\tNo. Model 2 iterations:  " << Model2_Iterations << '\n';
      }
      else if(attrib == "NO. ITERATIONS MODEL 3"){
	Model3_Iterations = atoi(attribval.c_str());
	cerr << "\tNo. Model 3 iterations:  " << Model3_Iterations <<'\n';
      }
      else if(attrib == "NO. ITERATIONS MODEL 4"){
	Model4_Iterations = atoi(attribval.c_str());
	cerr << "\tNo. Model 4 iterations:  " << Model4_Iterations <<'\n';
      }
      else if(attrib == "P0"){
	double temp = atof(attribval.c_str());
	if (temp < 1.0 && temp > 0.0){
	  P0 = temp ;
	  FIX_P0 = true ;
	  cerr << "\tp0 will be fixed to : " << P0 << '\n';
	}
	else {
	  cerr << " Invalid attribute value for P0 " << attribval << " ... ignoring it \n";
	}
      }
      else if(attrib == "PROB CUTOFF"){
	double temp = atof(attribval.c_str());
	if (temp < 1.0 && temp > 0.0){
	  PROB_CUTOFF = temp ;
	  cerr << "Probability cutoff threshold will be set to "<< PROB_CUTOFF << '\n';
	}
	else {
	  cerr << " Invalid attribute value for PROB CUTOFF " << attribval << " ... ignoring it \n";
	}
      }
      else if(attrib == "COUNTINCREASE CUTOFF"){
	double temp = atof(attribval.c_str());
	if (temp < 1.0 && temp > 0.0){
	  COUNTINCREASE_CUTOFF = temp ;
	  cerr << "Counts increment cutoff threshold will be set to "<< COUNTINCREASE_CUTOFF << '\n';
	}
	else {
	  cerr << " Invalid attribute value for COUNTINCREASE CUTOFF " << attribval << " ... ignoring it \n";
	}
      }
      else if(attrib == "PROB SMOOTH"){
	double temp = atof(attribval.c_str());
	if (temp < 1.0 && temp > 0.0){
	  PROB_SMOOTH = temp ;
	  cerr << "Probability smoothing (floor) value will be set to "<< PROB_CUTOFF << '\n';
	}
	else {
	  cerr << " Invalid attribute value for PROB SMOOTH " << attribval << " ... ignoring it \n";
	}
      }
      else if(attrib ==  "TRANSFER 2 TO 3? (Y/N)"){
	Transfer2to3 = (attribval == "Y" || attribval == "y" ||
			attribval == "YES" || attribval == "yes"); 
	cerr << "\tRun transfer from Model 2 to Model 3?  " << 
	  ((Transfer2to3)? "Yes\n" : "No\n") ;
      }
      else if(attrib ==  "DO PEGGING? (Y/N)"){
	Peg = (attribval == "Y" || attribval == "y" ||
	       attribval == "YES" || attribval == "yes"); 
	cerr << "PEGGING :  " << 
	  ((Peg)? "Yes\n" : "No\n") ;
      }
      else if(attrib ==  "NO FILE DUMPS? (Y/N)"){
	NODUMPS = (attribval == "Y" || attribval == "y" ||
	       attribval == "YES" || attribval == "yes"); 
	cerr << "NO FILE DUMPS :  " << 
	  ((NODUMPS)? "Yes\n" : "No\n") ;
      }
      else if(attrib ==  "SIMPLE TRANSFER? (Y/N)"){	
	if (attribval == "Y" || attribval == "y" || 
	    attribval == "YES" || attribval == "yes"){
	  Transfer = TRANSFER_SIMPLE ;
	  cerr << "\tModel 2 to Model3 tranfer uses simple transfer\n";
	}
      }
      else if(attrib == "AD TABLE SPECIAL BACKOFF"){
	int temp = atoi(attribval.c_str());
	switch(temp)
	  {
	  case 0:
	    break;
	  case 1:
	    ADTableIgnoreL=1;
	    break;
	  case 2:
	    ADTableIgnoreM=1;
	    break;
	  case 3:
	    ADTableIgnoreL=ADTableIgnoreM=1;
	    break;
	  default:
	    if( temp>=4&&temp<=6 )
	      ADTableSpecial=temp-3;
	    else
	      cerr << "Wrong value for option -adBackOff , ignoring it!.\n";
	    break;
	  }
	
	cerr << "\tAD TABLE SPECIAL BACKOFF has value:  " << temp << '\n';
      }
      else if(attrib == "MODEL 1 DUMP FREQUENCY"){
	Model1_Dump_Freq = atoi(attribval.c_str());
	cerr << "\tModel 1 table dump frequency:  " << Model1_Dump_Freq
	     << '\n';
      }
      else if(attrib == "MODEL 2 DUMP FREQUENCY"){
	Model2_Dump_Freq = atoi(attribval.c_str());
	cerr << "\tModel 2 table dump frequency:  " << Model2_Dump_Freq
	     << '\n';
      }
      else if(attrib == "TRANSFER DUMP FREQUENCY"){
	Transfer_Dump_Freq = atoi(attribval.c_str());
	cerr << "\tTransfer table dump frequency:  " << Transfer_Dump_Freq
	     << '\n';
      }
      else if(attrib == "MODEL 3 DUMP FREQUENCY"){
	Model3_Dump_Freq = atoi(attribval.c_str());
	cerr << "\tModel 3 table dump frequency:  " << Model3_Dump_Freq
	     << '\n';
      }
      else if(attrib == "MODEL 4 DUMP FREQUENCY"){
	Model4_Dump_Freq = atoi(attribval.c_str());
	cerr << "\tModel 4 table dump frequency:  " << Model4_Dump_Freq
	     << '\n';
      }
      else if(attrib == "LOG FILE"){
	LogFilename = attribval;
	cerr << "\tLog file:  " << LogFilename << '\n';
      }
      else if(attrib == "OUTPUT PATH"){
	if(attribval[attribval.size() - 1] != '/')
	  attribval += "/";
	OPath = attribval;
	cerr << "\tOutput path:  " << OPath << '\n';
      }
      else if(attrib == "OUTPUT FILE PREFIX"){
	Prefix = attribval ;
	cerr << "\tOutput file prefix:  " << Prefix << '\n';
      }
      else if(attrib == "LOG PROB. PATH"){
	if(attribval[attribval.size() - 1] != '/')
	  attribval += "/";
	LogProbPath = attribval;
	cerr << "\tLog probability table will be read from path:  " << 
	  LogProbPath << '\n';
      }
      else if(attrib == "t FILE"){
	t_Filename = attribval;
	cerr << "\tt file:  " << t_Filename << '\n';
      }
      else if(attrib ==  "a FILE"){
	a_Filename = attribval;
	cerr << "\ta file:  " << a_Filename << '\n';
      }
      else if(attrib == "d FILE"){
	d_Filename = attribval;
	cerr << "\td file:  " << d_Filename << '\n';
      }
      else if(attrib == "n FILE"){
	n_Filename = attribval;
	cerr << "\tn file:  " << n_Filename << '\n';
      }
      else if(attrib == "p0 FILE"){
	p0_Filename = attribval;
	cerr << "\tp0 file:  " << p0_Filename << '\n';
      }
      else if(attrib == "DICTIONARY"){
	dictionary_Filename = attribval;
	if(dictionary_Filename != ""){
	  cerr << "\tdictionary file:  " << dictionary_Filename << '\n';
	  useDict = true;
	}
      }
      else if ( line == ""){}
      else {
	cerr << "Unrecognized attribute :" << attrib << '\n';
      }
    }
  }
}


void parseArguments(int argc, char *argv[])
{

  int arg = 1, temp ;

  bool noconcat = false  ;
  if(!strcmp(argv[1], "--h") || !strcmp(argv[1], "--help")){
    printHelp();
    exit(1);
  }
  // NAS:  This is where we read the configuration file.  7/9
  parseConfigFile(argv[1]);
  cerr << "Done parsing configuration file\n";
  // go through command line arguments and overwrite those taken from config
  // file if necessary
  while(++arg<argc){
    if(!strcmp(argv[arg], "-S"))
      if(++arg < argc)
	SourceVocabFilename = argv[arg];
      else{
	--arg;
	cerr << "Warning: source vocabulary file name expected after option -S\nUsing filename from config file:  " << SourceVocabFilename << '\n';
      }
    else if(!strcmp(argv[arg], "-T"))
      if(++arg < argc)
	TargetVocabFilename = argv[arg];
      else{
	--arg;
	cerr << "Warning: target vocabulary file name expected after option -T\nUsing filename from config file:  " << TargetVocabFilename << '\n';
      }
    else if(!strcmp(argv[arg], "-C"))
      if(++arg < argc)
	CorpusFilename = argv[arg];
      else{
	--arg;
	cerr << "Warning: training corpus file name expected after option -C\nUsing filename from config file:  " << CorpusFilename << '\n';
      }
    else if(!strcmp(argv[arg], "-TC"))
      if(++arg < argc)
	TestCorpusFilename = argv[arg];
      else{
	--arg;
	cerr << "Warning: test corpus file name expected after option -TC\nUsing filename from config file:  " << TestCorpusFilename << '\n';
      }
    else if(!strcmp(argv[arg], "-d"))
      if(++arg < argc)
	dictionary_Filename = argv[arg];
      else{
	--arg;
	cerr << "Warning: dictionary filename expected after option -d , this option will be ignored\n";
      }
    else if (!strcmp(argv[arg], "--v"))
      Verbose = true;
    else if (!strcmp(argv[arg], "--NODUMPS"))
      NODUMPS = true;
    else if(!strcmp(argv[arg], "--h") || !strcmp(argv[arg], "--help")){
      printHelp();
      exit(0);
    }
    else if (!strcmp(argv[arg], "--p"))
      Peg = true;
    else if (!strcmp(argv[arg], "--st"))
      Transfer = TRANSFER_SIMPLE;
    else if (!strcmp(argv[arg], "-l")){
      Log = true ;
      if (++arg < argc)
	LogFilename = argv[arg];
      else{
	--arg;
	cerr << "Warning: log file name expected after option -l\nUsing filename:  "<< LogFilename <<'\n';
      }
    }
    else if (!strcmp(argv[arg], "-p0")){
      if (++arg < argc && atof(argv[arg]) > 0 && atof(argv[arg]) < 1.0){
	P0 = atof(argv[arg]);
	FIX_P0 = true ;
      }
      else{
	--arg;
	cerr << "Invalid attribute value after \"-p0\". A valid probability expected, ignoring it!\n";
      }
    }
    else if (!strcmp(argv[arg], "-probCutoff")){
      if (++arg < argc && atof(argv[arg]) > 0 && atof(argv[arg]) < 1.0){
	PROB_CUTOFF = atof(argv[arg]);
	cerr << "Probability cutoff threshold will be set to "<< PROB_CUTOFF << '\n';
      }
      else{
	--arg;
	cerr << "Invalid attribute value after \"-probCutoff\". A valid probability expected, ignoring it!\n";
      }
    }
    else if (!strcmp(argv[arg], "-countCutoff")){
      if (++arg < argc && atof(argv[arg]) > 0 && atof(argv[arg]) < 1.0){
	COUNTINCREASE_CUTOFF = atof(argv[arg]);
	cerr << "Counts increment cutoff threshold will be set to "<< COUNTINCREASE_CUTOFF << '\n';
      }
      else{
	--arg;
	cerr << "Invalid attribute value after \"-countCutoff\". A valid probability expected, ignoring it!\n";
      }
    }
    else if (!strcmp(argv[arg], "-probSmooth")){
      if (++arg < argc && atof(argv[arg]) > 0 && atof(argv[arg]) < 1.0){
	PROB_SMOOTH = atof(argv[arg]);
	cerr << "Probability smoothing (floor) value will be set to "<< PROB_CUTOFF << '\n';
      }
      else{
	--arg;
	cerr << "Invalid attribute value after \"-probSmooth\". A valid probability expected, ignoring it!\n";
      }
    }
    else if (!strcmp(argv[arg], "-adBackOff")){
      if (++arg < argc)
	{
	  int f=atoi(argv[arg]);
	  switch(f)
	    {
	    case 0:
	      break;
	    case 1:
	      ADTableIgnoreL=1;
	      break;
	    case 2:
	      ADTableIgnoreM=1;
	      break;
	    case 3:
	      ADTableIgnoreL=ADTableIgnoreM=1;
	      break;
	    default:
	      if( f>=4&&f<=6 )
		ADTableSpecial=f-3;
	      else
		cerr << "Wrong value for -adBackOff.\n";
	      break;
	    }
	}
      else{
	--arg;
	cerr << "Warning: missing value for option -adBackOff\n";
      }
    }
    else if (!strcmp(argv[arg], "-ml")){
      unsigned int temp ;
      if (++arg < argc){
	temp = atoi(argv[arg]) + 1;
	if (temp > 1 && temp <= MAX_SENTENCE_LENGTH_ALLOWED)
	  MAX_SENTENCE_LENGTH = temp ;
      }
      else{
	--arg;
	cerr << "Warning: max sentence length expected after option -ml\nUsing MAX SENTENCE LENGTH:  "<< MAX_SENTENCE_LENGTH <<'\n';
      }
    }
    else if (!strcmp(argv[arg], "-m1")){
      if (++arg < argc)
       Model1_Iterations = atoi(argv[arg]);
      else{
	--arg;
	cerr << "Warning: number of iterations expected after option -m1\nUsing Model1 Iterations:  "<< Model1_Iterations <<'\n';
      }
    }
    else if (!strcmp(argv[arg], "-m2to3")){
      Transfer2to3 = true ;
    }
    else if (!strcmp(argv[arg], "-m2")){
      if (++arg < argc)
       Model2_Iterations = atoi(argv[arg]);
      else{
	--arg;
	cerr << "Warning: number of iterations expected after option -m2\nUsing Model2 Iterations:  "<<Model2_Iterations<<'\n';
      }
    }
    else if (!strcmp(argv[arg], "-m3")){
      if (++arg < argc)
	Model3_Iterations = atoi(argv[arg]);
      else{
	--arg;
	cerr << "Warning: number of iterations expected after option -m3\nUsing Model3 Iterations:  "<<Model3_Iterations<<'\n';
      }
    }
    else if (!strcmp(argv[arg], "-m4")){
      if (++arg < argc)
	Model4_Iterations = atoi(argv[arg]);
      else{
	--arg;
	cerr << "Warning: number of iterations expected after option -m4\nUsing Model 4 Iterations:  "<<Model4_Iterations<<'\n';
      }
    }
    else if (!strcmp(argv[arg], "-t1")){
      if (++arg < argc){
	temp = atoi(argv[arg]);
	if (temp <= Model1_Iterations && temp >= 0)
	  Model1_Dump_Freq = temp ;
	else 
	  cerr << "Warning: invalid number after option -t1, dumping probability tables every " << Model1_Dump_Freq << " iteration(s) in model1 training\n"; 
      }
      else{
	--arg;
	cerr << "Warning: number missing after option -t1, dumping probability tables every " << Model1_Dump_Freq << " iteration(s) in Model1 training\n"; 
      }
    } /* end of if "-t1" */
    else if (!strcmp(argv[arg], "-t2")){
      if (++arg < argc){
	temp = atoi(argv[arg]);
	if (temp <= Model2_Iterations && temp >= 0)
	  Model2_Dump_Freq = temp ;
	else 
	  cerr << "Warning: invalid number after option -t2, dumping probability tables every " << Model2_Dump_Freq << " iteration(s) in model2 training\n"; 
      }
      else{
	--arg;
	cerr << "Warning: number missing after option -t2, dumping probability tables every " << Model2_Dump_Freq << " iteration(s) in Model2 training\n"; 
      }
    } /* end of if "-t2" */
    else if (!strcmp(argv[arg], "-t2to3")){
      if (++arg < argc){
	temp = atoi(argv[arg]);
	if (temp >= 0)
	  Transfer_Dump_Freq = temp ;
	else {
	  cerr << "Warning: invalid number after option -t2to3, probability tables will " ;
	  if (Transfer_Dump_Freq == 0)
	    cerr << "NOT ";
	  cerr << "be dumped after Transfer\n"; 
	}
      } 
      else{
	--arg;
	  cerr << "Warning: number missing after option -t2to3, probability tables will " ;
	  if (Transfer_Dump_Freq == 0)
	    cerr << "NOT ";
	  cerr << "be dumped after Transfer\n"; 
      }
    } /* end of if "-t2to3" */
    else if (!strcmp(argv[arg], "-t3")){
      if (++arg < argc){
	temp = atoi(argv[arg]);
	if (temp <= Model3_Iterations && temp >= 0)
	  Model3_Dump_Freq = temp ;
	else 
	  cerr << "Warning: invalid number after option -t3, dumping probability tables every " << Model3_Dump_Freq << " iteration(s) in model3 training\n"; 
      }
      else{
	--arg;
	cerr << "Warning: number missing after option -t3, dumping probability tables every " << Model3_Dump_Freq << " iteration(s) in Model3 training\n"; 
      }
    } /* end of if "-t3" */
    else if (!strcmp(argv[arg], "-t4")){
      if (++arg < argc){
	temp = atoi(argv[arg]);
	if (temp <= Model4_Iterations && temp >= 0)
	  Model4_Dump_Freq = temp ;
	else 
	  cerr << "Warning: invalid number after option -t4, dumping probability tables every " << Model4_Dump_Freq << " iteration(s) in model4 training\n"; 
      }
      else{
	--arg;
	cerr << "Warning: number missing after option -t4, dumping probability tables every " << Model4_Dump_Freq << " iteration(s) in Model4 training\n"; 
      }
    } /* end of if "-t4" */
    else if (!strcmp(argv[arg], "-o"))
      if (++arg < argc){
	noconcat = true ;
	Prefix = argv[arg];
      }
      else{
	--arg;
	cerr << "Warning: output files prefix expected after option -o\nUsing default prefix:  "<<Prefix<<'\n';
      }
    else 
      cerr << "Warning: ignoring unrecognized option:  "<< argv[arg] << '\n' ;  
  }

  /* NAS:  Need to make sure that everything being run is consecutive (i.e., can't run M1 and M3 without M2  and the transfer)!  7/10 */
  if(Model1_Iterations > 0 && (Model3_Iterations > 0 || Transfer2to3 || Model4_Iterations > 0)
     && Model2_Iterations == 0){
    cerr << "You wanted to run Model 1 and Model 3 iterations without Model 2?  Changing to " << Model2_Iterations << " Model 2 iterations." << '\n';
    Model2_Iterations = MAX_ITERATIONS_MODEL2;
  }
  if(Model2_Iterations > 0 && (Model3_Iterations > 0 || Model4_Iterations  > 0) && !Transfer2to3){
    cerr << "You wanted to run Model 2 and Model 3 iterations without the transfer?  Changing transfer flag to TRUE." << '\n';
    Transfer2to3 = true;
  }

  /* NAS:  Need to check for filenames of table files if not starting from the beginning!  7/10 
    Yaser extended to a more robust checks and added Model4 stuff */
  if(Model1_Iterations == 0 &&  (Model4_Iterations > 0 || Model2_Iterations > 0 || Model3_Iterations > 0 || Transfer2to3) && 
     !strcmp(t_Filename.c_str(), "")){
    cerr << "ERROR:  Must give filename for T table in the configuration file if you are skipping Model 1 training." << '\n';
    exit(1);
  }
  if(Model2_Iterations == 0 && (Model3_Iterations > 0 || Model4_Iterations > 0 || Transfer2to3) && 
     !strcmp(a_Filename.c_str(), "")){
    cerr << "ERROR:  Must give filename for A table in the configuration file if you are skipping Model 2 training." << '\n';
    exit(1);
  }
  if(!Transfer2to3 && (Model3_Iterations > 0 || Model4_Iterations > 0) && !strcmp(n_Filename.c_str(), "")){
    cerr << "ERROR:  Must give filename for N table in the configuration file if you are skipping Model 2 to 3 transfer." << '\n';
    exit(1);
  }
  if(!Transfer2to3 && (Model3_Iterations > 0 || Model4_Iterations > 0) && !strcmp(d_Filename.c_str(), "")){
    cerr << "ERROR:  Must give filename for D table in the configuration file if you are skipping Model 2 to 3 transfer." << '\n';
    exit(1);
  }
  if(!Transfer2to3 && (Model3_Iterations > 0 || Model4_Iterations > 0) && !strcmp(p0_Filename.c_str(), "")){
    cerr << "ERROR:  Must give filename for P0 value in the configuration file if you are skipping Model 2 to 3 transfer." << '\n';
    exit(1);
  }
  if (!noconcat)
    Prefix = (OPath + Prefix);
  LogFilename = (OPath + LogFilename);
}

