/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include <strstream>
#include <time.h>
#include <set.h>
#include "defs.h"
#include "vocab.h" 
#include "Perplexity.h"
#include "getSentence.h"
#include "TTables.h"
#include "Globals.h"

void printHelp(void)
{
  cerr << "Usage:\n\n" << Usage << '\n';
  cerr << "Options (these override parameters set in the config file):\n\n";
  cerr << "\t-S <source vocab filename>\tuse <source vocab filename> for source vocabulary.\n";
  cerr << "\t-T <target vocab filename>\tuse <target vocab filename> for target vocabulary.\n";
  cerr << "\t-C <corpus filename>\tuse <corpus filename> as the train parallel corpus.\n";
  cerr << "\t-TC <corpus filename>\tuse <corpus filename> as the test parallel corpus.\n";
  cerr << "\t-d <dictionary filename>\tif you are using a dictionary to constraint count collection for iterations 1 of model1 training.\n";
  cerr << "\t--v \t\t print verbose message, Warning this is not very descriptive and not systematic.\n";
  cerr << "\t--NODUMPS \t Do not write any files to disk (This will over write dump frequency options).\n";
  cerr << "\t--h[elp]\t\tprint this help\n";
  cerr << "\t--p\t\tUse pegging when generating alignments for Model3 training.  (Default NO PEGGING)\n";
  cerr << "\t--st\t\tto use a fixed ditribution for the fertility parameters when tranfering from model 2 to model 3 (Default complicated estimation)\n";
  cerr << "\t-l <logfilename>\tgenerate log file logfilename.  (Default:  FALSE; file <Prefix>.Log)\n";
  cerr << "\t-p0 <n>\t a probability for which p0 will be tied (Default P0 value will float). \n" ;
  cerr << "\t-probCutoff <n>\t a probability threshold where values below it will be set to the floor value (which could be set using \"-probSmooth\").\n";
  cerr << "\t-countCutoff <n>\t a threshold for which count values below it will be ignored\n";
  cerr << "\t-probSmooth <n>\t a probability smoothing value (floor) where not existing probability values will be set to it.\n";
  cerr << "\t-adBackOff <n>\t BackOff scheme for alignment and distortion tables. Valid values are:\n";
  cerr << "\t\t\t0: no backing off scheme will be used (4-D table).\n";
  cerr << "\t\t\t1: for ignoring source sentence length (3-D table).\n";
  cerr << "\t\t\t2: for ignoring target sentence length (3-D table).\n";
  cerr << "\t\t\t3: for ignoring both source and target sentence length (2-D table).\n";
  cerr << "\t\t\t4: for using the absolute value of the distortion distance and sentence length diffrenece (2-D table).\n";
  cerr << "\t\t\t5: for using distortion distance and sentence length diffrence (2-D table).\n";
  cerr << "\t\t\t6: for using distortion distance (3-D table).\n";
  cerr << "\t-m1 <n>\tnumber of training iterations for model 1. (Default " << MAX_ITERATIONS_MODEL1 << ")\n";
  cerr << "\t-m2 <n>\tnumber of training iterations for model 2. (Default " << MAX_ITERATIONS_MODEL2 << ")\n";
  cerr << "\t-m3 \twhether to transfer from Model2 to Model3 (Default transfer will" << ((Transfer2to3)? " ":" not " )<< "be done)\n";
  cerr << "\t-m3 <n>\tnumber of training iterations for model 3. (Default " << MAX_ITERATIONS_MODEL3 << ")\n";
  cerr << "\t-m4 <n>\tnumber of training iterations for model 4. (Default " << MAX_ITERATIONS_MODEL4 << ")\n";
  cerr << "\t-t1 <n>\thow frequent do you want to dump tables (1 mean every iterations) in Model1 training\n";
  cerr << "\t-t2 <n>\thow frequent do you want to dump tables (1 mean every iterations) in Model2 training\n";
  cerr << "\t-t2to3 <n>\twhether you want to dump tables (1 means files will be dumped) in Model2 to Model3 transfer\n";
  cerr << "\t-t3 <n>\thow frequent do you want to dump tables (1 mean every iterations) in Model3 training\n";
  cerr << "\t-t4 <n>\thow frequent do you want to dump tables (1 mean every iterations) in Model4 training\n";
  cerr << "\t-o <fileprefix>\tset output file path/prefix to fileprefix.  (Default is a timestamp combined with userid)\n";
  cerr << "\t-ml <n>\tset maximum sentence length to n. (Default " <<
    MAX_SENTENCE_LENGTH_ALLOWED << " )\n";
}


void printOptions(ostream& of)
{ 

  time_t st = time(NULL);
  of << "==========================================================\n";
  of << "Program Started at: "<< ctime(&st) <<'\n';
  of << "-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n";
  of << "Running with the following options:\n";
  of << "\tSource Vocabulary Filename:\t"<<SourceVocabFilename<<'\n';
  of << "\tTarget Vocabulary Filename:\t"<<TargetVocabFilename<<'\n';
  of << "\tTraining Corpus Filename:\t"<<CorpusFilename<<'\n';
  if (TestCorpusFilename != "")
    of << "\tTesting Corpus Filename:\t"<<TestCorpusFilename<<'\n';
  of << "\tOutput File Path/Prefix:\t"<<Prefix << '\n';
  of << "\tVERBOSE:\t\t" << ((Verbose)? "TRUE" : "FALSE") << '\n';
  of << "\tGenerate Log:\t\t" << ((Log)? "TRUE": "FALSE" ) << '\n';  
  if (Log)
    of << "\tLog Filename:\t\t"<<LogFilename << '\n';
  if (Peg)
    of << "\tModel3 training with pegging \n";
  else
    of << "\tModel3 training without pegging \n";  
  of << "\tNumber of Training Iterations for: \n";
  of << "\t\t\t Model1:\t" << Model1_Iterations << '\n';
  if(Model1_Iterations == 0 && Model2_Iterations > 0)
    of << "\t\t\t\tReading t table from:\t" << t_Filename << '\n';
  of << "\t\t\t Model2:\t" << Model2_Iterations << '\n';
  if(Model2_Iterations == 0 && Transfer2to3){
    of << "\t\t\t\tReading a table from:\t" << a_Filename << '\n';
  }
  of << "\t\t\t Transfer?\t" << ((Transfer2to3)? ((Transfer==TRANSFER_SIMPLE)?
						  "SIMPLE" : "NOT SIMPLE"):"NO")
     << '\n';
  if(!Transfer2to3 && Model3_Iterations > 0 || Model4_Iterations > 0){
    of << "\t\t\t\tReading d table from:\t" << d_Filename << '\n';
    of << "\t\t\t\tReading n table from:\t" << n_Filename << '\n';
    of << "\t\t\t\tReading p0 table from:\t" << p0_Filename << '\n';
  }
  of << "\t\t\t Model3:\t" << Model3_Iterations << '\n';
  of << "\t\t\t Model4:\t" << Model4_Iterations << '\n';
  of << "\tFrequency of Dumping Probability Tables for: \n";
  if (NODUMPS)
    of << "\t\t\t No files will be dumped to disk!\n";
  else {
    of << "\t\t\t Model1:\t every " << Model1_Dump_Freq << " iteration(s)\n";
    of << "\t\t\t Model2:\t every " << Model2_Dump_Freq << " iteration(s)\n";
    if (Transfer_Dump_Freq == 1)
      of << "\t\t\t Transfer:\t will be dumped\n";
    else 
      of << "\t\t\t Transfer:\t will NOT be dumped\n";
    of << "\t\t\t Model3:\t every " << Model3_Dump_Freq << " iteration(s)\n";
    of << "\t\t\t Model4:\t every " << Model4_Dump_Freq << " iteration(s)\n";
  }
  of << "\tMAX SENTENCE LENGTH " << MAX_SENTENCE_LENGTH-1 << '\n';
  if (FIX_P0)
    of << "P0 value will be fixed to " << P0 << " during Model3 training \n";
  if (ADTableIgnoreL)
    of <<"The source sentence length dimension of (d&a) tables will be ignored\n";
  if (ADTableIgnoreM)
    of <<"The target sentence length dimension of (d&a) tables will be ignored\n";
  of << "-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*\n";
}


void generatePerplexityReport(const Perplexity& trainperp, 
			      const Perplexity& testperp, 
			      const Perplexity& trainVperp, 
			      const Perplexity& testVperp, 
			      ostream& of, int trainsize, int testsize, 
			      bool is_perp)
{
  unsigned int i, j, iter=0, m, n ;
  unsigned int last = min(trainperp.size(), testperp.size());
  unsigned int temp = min(trainVperp.size(), testVperp.size());
  last = min(last, temp);
  cerr << "last : " << last << '\n';
  if (is_perp)
    of << "# train-size test-size iter. model train-perplexity test-perplexity final(y/n) train-viterbi-perp test-viterbi-perp\n";
  else 
    of << "# train-size test-size iter. model train-cross-entropy test-cross-entropy final(y/n) train_viterbi_perp test_viterbi_perp\n";
  
  for (i = 0 ; i < last ; i++){
    if (is_perp){
      m = max(trainperp.perp[i].size(), testperp.perp[i].size());
      n = max(trainVperp.perp[i].size(), testVperp.perp[i].size());
    }
    else {
      m = max(trainperp.ce[i].size(), testperp.ce[i].size());
      n = max(trainVperp.ce[i].size(), testVperp.ce[i].size());
    }
    m = max(m, n) ;
    for (j = 0 ; j < m ; j++){
      bool flag = (j  == m-1 && i == last-1); // last element in report
      of << '\t' << trainsize << '\t' << testsize << '\t' << j+iter ;
      of << '\t' << trainperp.name[i] << '\t' ;
      if (is_perp){
	if (j < trainperp.perp[i].size())
	  of << trainperp.perp[i][j] << "\t\t" ;
	else 
	  of << "N/A\t\t";
	if (j < testperp.perp[i].size())
	  of << testperp.perp[i][j] << "\t\t" ;
	else 
	  of << "N/A\t\t";
      }
      else {
	if (j < trainperp.ce[i].size())
	  of << trainperp.ce[i][j] << "\t\t" ;
	else 
	  of << "N/A\t\t";
	if (j < testperp.ce[i].size())
	  of << testperp.ce[i][j] << "\t\t" ;
	else 
	  of << "N/A\t\t";
      }
      if (flag)
	of << "y\t" ;
      else of << "n\t";
      if (is_perp){
	if (j < trainVperp.perp[i].size())
	  of << trainVperp.perp[i][j] << "\t\t" ;
	else 
	  of << "N/A\t";
	if (j < testVperp.perp[i].size())
	  of << testVperp.perp[i][j] << '\n' ;
	else 
	  of << "N/A\n";
      }
      else {
	if (j < trainVperp.ce[i].size())
	  of << trainVperp.ce[i][j] << "\t\t" ;
	else 
	  of << "N/A\t\t";
	if (j < testVperp.ce[i].size())
	  of << testVperp.ce[i][j] << '\n' ;
	else 
	  of << "N/A\n";
      }
    }
    iter += j ;
  }
}

void  printSentencePair(Vector<WordIndex>& es, 
			Vector<WordIndex>& fs, 
			ostream& of)
  
  // just writes a sentece pair to the give output stream, one sentence pair line
  // it writes token ids not actual tokens.
{
  WordIndex i, j, l, m;
  l = es.size() - 1;
  m = fs.size() - 1;
  of << "Source sentence length : " << l << " , target : " << m << "\n"; 
  for (i = 1 ; i <= l ; i++)
    of << es[i] << ' ';
  of << "\n";
  for (j = 1 ; j <= m ; j++)
    of << fs[j] << ' ';
  of << "\n";

}

void printAlignToFile(const Vector<WordIndex>& es, 
		      const Vector<WordIndex>& fs, 
		      const Vector<WordEntry>& evlist, 
		      const Vector<WordEntry>& fvlist, 
		      ostream& of2, 
		      const Vector<WordIndex>& viterbi_alignment, 
		      int pair_no, double alignment_score)
     
     // prints the given alignment to alignments file (given it stream pointer)
     // in a format recognizable by the draw-alignment tool ... which is of the
     // example (each line triple is one sentence pair): 
     //   # sentence caption 
     //   target_word_1 target_word_2  ..... target_word_m
     //   source_word_1 ({ x y z }) source_word_2 ({ })  .. source_word_n ({w})
     // where x, y, z, and w are positions of target words that each source word
     // is connected to.

{
  WordIndex l, m;
  Vector<Vector<WordIndex> > translations(es.size()); // each english words has a vector 
  // of zero or more translations .
  l = es.size() - 1;
  m = fs.size() - 1;
  of2 << "# Sentence pair (" << pair_no <<") source length " << l << " target length "<< m << 
    " alignment score : "<< alignment_score << '\n';
  for (WordIndex j = 1 ; j <= m ; j++){
    of2 << fvlist[fs[j]].word << " " ;
    translations[viterbi_alignment[j]].push_back(j);
  }
  of2 << '\n';
  
  for (WordIndex i = 0  ; i <= l ; i++){ // these loops are not efficient 
    of2 << evlist[es[i]].word << " ({ " ;
    //    for (j = 1 ; j <= m ; j++)
    //      if (viterbi_alignment[j] == i){
    for (WordIndex j = 0 ; j < translations[i].size() ; j++)
      of2 << translations[i][j] << " " ;
    of2 << "}) ";
  }
  of2 << '\n';
}


void printOverlapReport(const tmodel<COUNT, PROB>& tTable, sentenceHandler& trainHandler, 
			sentenceHandler& testHandler,  vcbList& trainEList, 
			vcbList& trainFList, vcbList& testEList, vcbList& testFList)
{
  set<pair<WordIndex, WordIndex> > testCoocur ;
  sentPair s ;
  /*  string unseenCoocurFile = Prefix + ".tst.unseen.cooc" ;
      ofstream of_unseenCoocur(unseenCoocurFile.c_str());
      
      string seenCoocurFile = Prefix + ".tst.seen.cooc" ;
      ofstream of_seenCoocur(seenCoocurFile.c_str());
  */  
  testHandler.rewind();
  int seen_coocur = 0, unseen_coocur = 0, srcUnk = 0, trgUnk = 0 ;
  while(testHandler.getNextSentence(s)){    
    for (WordIndex i = 1 ; i < s.eSent.size() ; i++)
      for (WordIndex j = 1 ; j < s.fSent.size() ; j++)	
	testCoocur.insert(pair<WordIndex, WordIndex> (s.eSent[i], s.fSent[j])) ;
  }
  set<pair<WordIndex, WordIndex> >::const_iterator i ;
  for (i = testCoocur.begin() ; i != testCoocur.end() ; ++i){
    if (tTable.getProb((*i).first, (*i).second) > PROB_SMOOTH){
      seen_coocur ++ ;
      //      of_seenCoocur << (*i).first << ' ' << (*i).second << '\n';
    }
    else {
      unseen_coocur++;
      //      of_unseenCoocur << (*i).first << ' ' << (*i).second << '\n';
    }
  }
  
  string trgUnkFile = Prefix + ".tst.trg.unk" ;
  ofstream of_trgUnk(trgUnkFile.c_str());

  for (WordIndex i = 0 ; i <  testFList.getVocabList().size() && i < testFList.uniqTokens();i++)
    if (testFList.getVocabList()[i].freq > 0 && trainFList.getVocabList()[i].freq <= 0){
      of_trgUnk << i << ' ' << testFList.getVocabList()[i].word << ' ' << testFList.getVocabList()[i].freq
		<< '\n';
      trgUnk++ ;
    }
  string srcUnkFile = Prefix + ".tst.src.unk" ;
  ofstream of_srcUnk(srcUnkFile.c_str());

  for (WordIndex j = 0 ; j <  testEList.getVocabList().size() && j < testEList.uniqTokens();j++)
    if (testEList.getVocabList()[j].freq > 0 && trainEList.getVocabList()[j].freq <= 0){
      srcUnk++ ;
      of_srcUnk << j << ' ' << testEList.getVocabList()[j].word << ' ' << testEList.getVocabList()[j].freq
		<< '\n';
    }
  string summaryFile = Prefix + ".tst.stats" ;  
  ofstream of_summary(summaryFile.c_str());
  of_summary << "\t\t STATISTICS ABOUT TEST CORPUS\n\n";
  of_summary << "source unique tokens: " <<  testEList.uniqTokens() << '\n';
  of_summary << "target unique tokens: " <<  testFList.uniqTokens() << '\n';
  of_summary << "unique unseen source tokens: " << srcUnk << '\n';
  of_summary << "unique unseen target tokens: " << trgUnk << '\n';
  of_summary << "cooccurrences not found in the final t table: " << unseen_coocur << '\n';
  of_summary << "cooccurrences found in the final t table: " << seen_coocur << '\n';
  
}

