/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
transpair_model3: representation of a translation pair for model3 training
allowing for fast access (esp. to t table).

Franz Josef Och (30/07/99)
--*/
#include "transpair_model3.h"

transpair_model3::transpair_model3(const Vector<WordIndex>&es, const Vector<WordIndex>&fs, tmodel<COUNT, PROB>&tTable, amodel<PROB>&aTable, amodel<PROB>&dTable, nmodel<PROB>&nTable, double _p1, double _p0, void*)
  : t(es.size(), fs.size()), d(es.size(), fs.size()), a(es.size(), fs.size()), n(es.size(), MAX_FERTILITY+1)/*, e(es.size(), fs.size())*/, l(es.size()-1), m(fs.size()-1), p0(_p0), p1(_p1)
{
  //static const PROB zero=0.0;
  WordIndex l=es.size()-1,m=fs.size()-1;
  for(WordIndex i=0;i<=l;i++)
    {
      for(WordIndex j=1;j<=m;j++)
	{
	  t(i, j)=tTable.getProb(es[i], fs[j]);
	  d(i, j)=dTable.getValue(j, i, l, m);
	  a(i, j)=aTable.getValue(i, j, l, m);
	  //e(i, j)=(t(i, j)>PROB_SMOOTH&&d(i, j)>PROB_SMOOTH&&a(i, j)>PROB_SMOOTH);
	}
      for(WordIndex f=0;f<MAX_FERTILITY;f++)
	{
	  n(i, f)=nTable.getValue(es[i], f);
	}
      n(i,MAX_FERTILITY)=PROB_SMOOTH;
    }
}

ostream&operator<<(ostream&out, const transpair_model3&m)
{
  for(WordIndex i=0;i<=m.get_l();i++)
    {
      out << "EF-I:"<<i<<' ';
      for(WordIndex j=1;j<=m.get_m();j++)
	out << "("<<m.t(i, j)<<", "<<m.d(i, j)/*<<", "<<m.e(i, j)*/<<")";
      for(WordIndex j=1;j<MAX_FERTILITY;j++)
	out << "(fert:"<<m.get_fertility(i, j)<<")";
      out << endl;
    }
  out << "T:" << m.t << "D:" << m.d << "A:" << m.a  << "N:" << m.n << m.p0 << m.p1 << endl;
  return out;
}


//PROB transpair_model3::scoreOfMove(const alignment&a, int new_i, int j)const
LogProb transpair_model3::scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j, bool withDistortions)const
{
  LogProb change;
  const WordIndex old_i=a(j);
  WordIndex f0=a.fert(0);
  if (old_i == new_i)
    change=1.0;
  else if (old_i == 0)
    change=((double)p0*p0/p1) *
      ((f0*(m-f0+1.0)) / ((m-2*f0+1)*(m-2*f0+2.0))) *
      (PROB)(a.fert(new_i)+1) *
      (get_fertility(new_i, a.fert(new_i)+1) / get_fertility(new_i, a.fert(new_i)))*
      (t(new_i, j)/t(old_i, j))*
      (withDistortions?d(new_i, j):1.0);
  else if (new_i == 0)
    change=
      ((double(p1) / (p0*p0)) *
       (double((m-2*f0)*(m-2*f0-1))/((f0+1)*(m-f0))) *
       (1.0/a.fert(old_i)) *
       (get_fertility(old_i, a.fert(old_i)-1) /get_fertility(old_i, a.fert(old_i)))*
       (t(new_i, j) /t(old_i, j)) *
       (withDistortions?(1.0 / d(old_i, j)):1.0));
  else
    {
#ifdef DEBUG
      double t1=a.fert(old_i);
      double t2=get_fertility(old_i,a.fert(old_i));
      double t3=get_fertility(new_i,a.fert(new_i));
      double t4=t(old_i,j);
      if( t1==0 || t2==0 || t3==0 || t4==0 )
	cerr << t1 << " " << t2 << " " << t3 << " " << t4 << " " << old_i << " " << new_i << " " << a << " " << *this << endl;
#endif
      change=
	(double(a.fert(new_i)+1)/a.fert(old_i)) *
	(get_fertility(old_i,a.fert(old_i)-1) / get_fertility(old_i,a.fert(old_i))) *
	(get_fertility(new_i,a.fert(new_i)+1) /get_fertility(new_i,a.fert(new_i))) *
	(t(new_i,j)/t(old_i,j)) *
	(withDistortions?(d(new_i,j)/d(old_i,j)):1.0);
    }
#ifdef VDEBUG
  if( withDistortions )
    {
      LogProb realCosts=_scoreOfMove(a, new_i, j);
      if( (1.0-realCosts/change)>3e-2 )
	{
	  cerr << "Wrong costs of move: " << realCosts << " " << change << " " << (1.0-realCosts/change)<< endl;
	  cerr << a << endl;
	  cerr << new_i << " " << j << " " << old_i << endl;
	}
    }
#endif
  return change;
}
//PROB transpair_model3::scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2)const 
LogProb transpair_model3::scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2, bool withDistortions)const 
{
  PROB score;
  assert(j1<j2);
  WordIndex i1=a(j1), i2=a(j2);
  if (i1==i2) 
    score=1;
  else
    {
      score=(t(i2, j1)/t(i1, j1))*(t(i1, j2)/t(i2, j2));
      if( withDistortions )
	{
	  if (i1)
	    score *= d(i1, j2)/d(i1, j1);
	  if (i2)
	    score *= d(i2, j1)/d(i2, j2);    
	}
    }
#ifdef VDEBUG
  if( withDistortions )
    {
      PROB realCosts=_scoreOfSwap(a, j1, j2);
      if( (1.0-realCosts/score)>3e-2 )
	{
	  cerr << "Wrong costs of swap: " << realCosts << " " << score << " " << (1.0-realCosts/score) << endl;
	  cerr << a <<endl;
	}
    }
#endif
  return score;
}
