#!/usr/local/perl

# whittle.pl
# ==========
# Mike Jahr, 7/22/99
# Machine Translation group, WS99
# Center for Language and Speech Processing
# 
# perl script designed to generate parallel corpus files suitable
# for a candide-style mt trainer. also separates data into training
# and test sets.
#
# warning: since it hashes all sentences to check for duplicates,
# whittle requires a very large chunk of ram to run on large corpora.
# (this doesn't apply to the --split option)
#
# for help run with the "-h" switch or see options listed below.

$VERSION = "whittle v1.0, by mike jahr, 8/13/99";
$USAGE = "usage: ".file_name($0)." [options] <input-directory | input-stem>";
$OPTIONS = <<EOT;
   -a       all: do not split into training and test sets
  *-b nn    specify baseline test set size; nn can be either an integer
              (the desired number of sentences) or a decimal (fraction of
              the total corpus size)
   -c       only generate a counts/vocab file; usually done automatically
   -d dd    specify output file directory; overriden by -o
   -E xx    specify suffix for English files
   -f i     specify minimum word frequency; less-frequent words will be
              replaced with the UNKNOWN token. i<=1 retains all words.
   -F xx    specify suffix for foreign-language files
   -h       print this help information
  *-i nn    interleaving ratio: fraction of test set to be spread evenly
              throughout the corpus. the rest is taken from the end.
   -l i     specify maximum sentence length; pairs where one or both are
              longer will be discarded. use i=0 for no max length.
   -n i     assign word identifiers starting from i
   -o d/p   specify output file directory and prefix; overrides -d
   -s nn    specify total output corpus size; nn as above
   -t nn    specify training set size; nn as above
   -T nn    specify test set size; nn as above. note that the test set
              cannot be larger than specified by the -b value.
   -u i     specify identifier of unknown word
   -v       print version  information
   -w       word format, no integerizing; output in human-readable form
  --all         do not split corpus; identical to -a
  --bst         output in binary (.bst) format
  --snt         output in .snt format, one sentence per line
  --sfreq       sort output by sentence frequency
  --split       only split the corpora: output parallel files in the
                  same format as the input
  --test nn     specify test set size; identical to -T
  --test-only   only generate a test set; equivalent to --train 0
  --train nn    specify training set size; identical to -t
  --train-only  only generate a training set; equivalent to --test 0

NOTE: Switches marked with a * change how sentences will be divided into
      training and test sets. -T and -t do not have this effect. To 
      guarantee mutually exclusive training and test corpora across
      multiple runs of whittle, use the same -b and -i values over all
      runs, or first split the corpus using --split and then run
      using --all on the individual corpora.
EOT

# first init globals with default values

$MAX_LENGTH = undef; # max sentence length
$MIN_FREQ = undef;   # minimum word frequency
$UNKNOWN_WORD = "UNK"; # unknown word placeholder
$UNKNOWN_ID = 1;     # id of unknown word
$FIRST_WORD_ID = 2;  # word numbering starts at 2
$FORMAT = ".snt";    # snt is default format
$INTEGERIZE = 1;     # integerizing is on
$BASE_RATIO = 0.02;  # test set is 2% of corpus
$CORPUS_PROP = 1;    # use entire corpus
$TEST_PROP = $CORPUS_PROP;
$TRAIN_PROP = $CORPUS_PROP;
$INT_PROP = 0.5;     # half of test sentences interleaved
$ENG_SUFFIX = "_e";  # suffix for english files
$FOR_SUFFIX = "_f";  # suffix for french files
$COUNT_SUFFIX = ".cnt";
$VOCAB_SUFFIX = ".vcb";
$OUT_DIR = undef;    # output goes in the input directory
$TEMP_PREFIX = "whittle";
$TEMP_SUFFIX = ".temp";
$NA = "N/A"; # used for training/test size in --all
  
init();

if ($COUNT_ONLY) {
  count();
} else {
  count_load();
  
  print "\n==============================================\n";
  print "Generating corpora\n\n";
  
  # now calculate values for corpus sizes and ratios
  # for explanations see sub print_corp_info
  $base_test_size = ($BASE_RATIO > 1) ? 
    $BASE_RATIO : $BASE_RATIO*$NUM_SENT;
  $base_train_size = $NUM_SENT - $base_test_size;
  $corp_size = ($CORPUS_PROP > 1) ?
    $CORPUS_PROP : ($CORPUS_PROP*$NUM_SENT);
  $CORPUS_PROP /= $NUM_SENT if ($CORPUS_PROP > 1);
  ($corp_size && $CORPUS_PROP) or
    die "Illegal corpus size. whittle aborting.\n";
  
  $test_size = ($TEST_PROP > 1) ?
    $TEST_PROP : $TEST_PROP*$CORPUS_PROP*$base_test_size;
  $test_size = $base_test_size if ($test_size > $base_test_size);
  if ($test_size != 0) {
    $test_ratio = ((int($base_test_size) == int($test_size)) ?
		   undef : $base_test_size / $test_size);
  } else { $TRAIN_ONLY = 1; }
  
  $train_size = ($TRAIN_PROP > 1) ?
    $TRAIN_PROP : $TRAIN_PROP*$CORPUS_PROP*$base_train_size;
  $train_size = $corp_size - $test_size
    if ($CUSTOM_TEST && !$CUSTOM_TRAIN);
  if ($train_size > $base_train_size) {
    $train_size = $base_train_size;
    $CUSTOM_TRAIN = 1;
  }
  if ($train_size != 0) {
    $train_ratio = ((int($base_train_size) == int($train_size)) ?
		    undef : $base_train_size / $train_size);
  } else { $TEST_ONLY = 1; }

  if ($CUSTOM_TRAIN && !$CUSTOM_TEST) {
    $test_size = $corp_size - $train_size;
    if ($test_size > $base_test_size) {
      $test_size = $base_test_size;
      $CUSTOM_TEST = 1;
    }
    if ($test_size != 0) {
      $test_ratio = ((int($base_test_size) == int($test_size)) ?
		     undef : $base_test_size / $test_size);
    } else { $TRAIN_ONLY = 1; }
  }

  $corp_size = $test_size + $train_size 
    if ($CUSTOM_TEST && $CUSTOM_TRAIN);
  
  $int_test_size = ($INT_PROP > 1) ?
    $INT_PROP : $base_test_size*$INT_PROP;
  $end_test_size = $base_test_size - $int_test_size;
  $last_train = $NUM_SENT - $end_test_size;
  # $train_ratio = $NUM_SENT / $corp_size;
  if ($int_test_size != 0) {
    $int_ratio = ($NUM_SENT - $end_test_size) / $int_test_size;
  } else { $int_ratio = $last_train; }
  
  print_corp_info();
  
  $train_c = $test_c = $train_discard = $test_discard = 0;
  $pair_no = $test_no = $train_no = 0;
  
  $next_train_used = $train_ratio/2 if $train_ratio;
  $next_test_used = $test_ratio/2 if $test_ratio;
  
  $next_test_pair = $int_ratio/2;
  
  # now make temp files
  if ($FORMAT) {
    unless ($TEST_ONLY) {
      if ($NOSPLIT) {
	$t = temp_name("");
      } else {
	$t = temp_name("train");
      }
      print "creating temporary file $t\n";
      open(TRAIN, ">$t") || die "couldn't create temporary file $t\n";
    }
    unless ($TRAIN_ONLY) {
      $t = temp_name("test");
      print "creating temporary file $t\n";
      open(TEST, ">$t") || die "couldn't create temporary file $t\n";
    }
  } else { # we're only splitting
    unless ($TEST_ONLY) {
      $t = $OUT_DIR.$OUTFILE;
      $t .= "-train" unless $NOSPLIT;
      open(ETRAIN, ">$t$ENG_SUFFIX") 
	|| die "couldn't write to file $t$ENG_SUFFIX\n";
      open(FTRAIN, ">$t$FOR_SUFFIX") 
	|| die "couldn't write to file $t$FOR_SUFFIX\n";
    }
    unless ($TRAIN_ONLY) {
      $t = "$OUT_DIR$OUTFILE-test-int".$ENG_SUFFIX;
      open(ETEST, ">$t") || die "couldn't write to file $t\n";
      $t = "$OUT_DIR$OUTFILE-test-int".$FOR_SUFFIX;
      open(FTEST, ">$t") || die "couldn't write to file $t\n";
    }
  }
  
#  print "last_train: $last_train\n";
#  print "next_test_used: $next_test_used\n";
#  print "next_test_pair: $next_test_pair\n";
#  print "next_train_used: $next_train_used\n";
  
  for ($i = 0; $i <= $#ENG_FILES; $i++) {
    $e = $ENG_FILES[$i];
    $f = $FOR_FILES[$i];
    open(E, "<$e") || die "Couldn't read file <$e>\n";
    open(F, "<$f") || die "Couldn't read file <$f>\n";
    print "processing " . file_name($e) . " & " .
      file_name($f) . "\n";
    while ($e = <E>, $f = <F>) {
      unless ($e && $f) {
	warn "WARNING: files $e and $f appear to have different " . 
	  "numbers of sentences...\n";
	last;
      }
      if (++$pair_no >= $next_test_pair) { # it is a test sent
	if ($next_test_pair < $last_train) {
	  if (($next_test_pair += $int_ratio) > $last_train) {
	    $next_test_pair = $last_train+1;
	    #print "+next_test_pair: $next_test_pair\n";
	    unless ($FORMAT || $TRAIN_ONLY) {
	      close(ETEST) && close(FTEST)
		|| warn "couldn't close test outfiles\n";
	      $t = "$OUT_DIR$OUTFILE-test-end".$ENG_SUFFIX;
	      open(ETEST, ">$t") || die "couldn't write to file $t\n";
	      $t = "$OUT_DIR$OUTFILE-test-end".$FOR_SUFFIX;
	      open(FTEST, ">$t") || die "couldn't write to file $t\n";
	    }
	  }
	}
	if (!$test_ratio || ++$test_no > $next_test_used) {
	  if (!$TRAIN_ONLY) {
	    $pair = process(\$e, \$f);
	    if ($pair) {
	      $test_c++;
	      $next_test_used += $test_ratio if ($test_ratio);
	      if ($FORMAT) {
		print TEST "$$pair"
		  if (++$test_counts{$$pair} == 1);
	      } else {
		($e, $f) = split(/\n/, $$pair) if ($MIN_FREQ);

		print ETEST $e."\n";
		print FTEST $f."\n";
	      }
	    } else {
	      #print "******discarding test\n";
	      $test_discard++;
	    }
	  }
	}
	#print "*test-no: $test_no; next: $next_test_used; c: $test_c\n";
      } else {
	if (!$TEST_ONLY && (!$train_ratio || ++$train_no >= $next_train_used)) {
#	print "next_train_used: $next_train_used\n";
#	print "train_c: $train_c\n";
	$pair = process(\$e, \$f);
	if ($pair) {
	  $train_c++;
	  $next_train_used += $train_ratio if ($train_ratio);
	  if ($FORMAT) {
	    print TRAIN "$$pair"
	      if (++$train_counts{$$pair} == 1);
	  } else {
	    ($e, $f) = split(/\n/, $$pair) if ($MIN_FREQ);
#	      $pair =~ /([^\n]+\n)([^\n]+\n)/;
	    print ETRAIN $e."\n";
	    print FTRAIN $f."\n";
	  }
	} else {
	  #print "-----discarding train\n";
	  $train_discard++;
	}
      }
	#print "-train-no: $train_no; next: $next_train_used; c: $train_c\n";
      }
    }
    close E;
    close F;
  }
  
  print "==============================================\n";
  print "Done...made $train_c training, $test_c test.\n";
  print "skipped $train_discard potential training and " .
    "$test_discard potential\n  " . 
    "test sentence pairs more than $MAX_LENGTH words long.\n"
      if ($MAX_LENGTH);

  print "\nwriting corpus files to disk...\n";
  write_corp_stats($train_c, $test_c);

  if ($FORMAT) {
    unless ($TEST_ONLY) {
      close(TRAIN) || warn "couldn't close temp training file.\n";
      $t = "";
      $t = "train" unless $NOSPLIT;
      write_snt($t, \%train_counts) if $SNT;
      write_bst($t, \%train_counts) if $BST;
    }
    unless ($TRAIN_ONLY) {
      close(TEST) || warn "couldn't close temp test file.\n";
      write_snt("test", \%test_counts) if $SNT;
      write_bst("test", \%test_counts) if $BST;
    }
  } else {
    close(ETEST) && close(FTEST) && close(ETRAIN) && close(FTRAIN)
      || warn "couldn't close outfiles\n";
  }
}

@times = times;
$utime = norm_time($times[0]);
$stime = norm_time($times[1]);
print "\nwhittle done. elapsed time: $stime system, $utime user.\n";

########### SUBROUTINES
#takes in seconds, outputs hour:min:sec
sub norm_time {
  my $s = shift;
  return $s."s" if $s < 60;
  my $h = int($s/3600);
  $s -= $h*3600;
  my $m = int($s/60);
  $s -= $m*60;
  $s = int($s+100)/100;
  $s = "0$s" if $s < 10;
  return "$m:$s" unless $h;
  $m = "0$m" if $m < 10;
  return "$h:$m:$s";
}

# initialize whittle
sub init {
  if ($ARGV[0] && $ARGV[0] eq "-h") {
    print "$VERSION\n\n$USAGE\n$OPTIONS\n";
    print "Default options:\n";
    print_options();
    die "\n";
  } elsif ($ARGV[0] && $ARGV[0] eq "-v") {
    die "$VERSION\n";
  }
  print "Welcome to $VERSION\n";
  print "For help use the -h switch\n" if (!$ARGV[0]);
  
  # parse command-line arguments
  while ($ARGV[0] && $ARGV[0] =~ /^-/) {
    $_ = shift @ARGV;
    #if (/^-D(.*)/) { $debug = $1 }
    if (/^--wit$/)     { die "sorry, wit format not yet implemented.\n";
			 $FORMAT = ".wit";             }
    elsif (/^--snt$/)  { $FORMAT = ".snt";             }
    elsif (/^-a$/ || 
	   /^--all$/)  { $NOSPLIT = 1;                 }
    elsif (/^--sfreq$/){ die "sorry, sfreq not yet implemented";
		         $SFREQ = 1;                   }
    elsif (/^--split$/){ $FORMAT = undef;              }
    elsif (/^--bst$/)  { die "sorry, binary format not yet implemented.\n";
			 $FORMAT = ".bst";             }
    elsif (/^-w$/)     { $INTEGERIZE = 0;              }
    elsif (/^-c$/)     { $COUNT_ONLY = 1;              }
    elsif (/^--test-only$/) { $TEST_ONLY = 1;          }
    elsif (/^--train-only$/){ $TRAIN_ONLY = 1;         }
    elsif (/^-b$/)     { $BASE_RATIO = shift @ARGV;    }
    elsif (/^-f$/)     { $MIN_FREQ = shift @ARGV;      }
    elsif (/^-l$/)     { $MAX_LENGTH = shift @ARGV;    }
    elsif (/^-u$/)     { $UNKNOWN_ID = shift @ARGV;    }
    elsif (/^-s$/)     { $CORPUS_PROP = shift @ARGV;   }
                         #$CUSTOM_CORPUS = 1;          }
    elsif (/^-n$/)     { $FIRST_WORD_ID = shift @ARGV; }
    elsif (/^-d$/)     { $OUT_DIR = shift @ARGV;       }
    elsif (/^-o$/)     { $OUTFILE = shift @ARGV;       }
    elsif (/^-t$/ || 
	   /^--train$/){ $TRAIN_PROP = shift @ARGV; 
			 $CUSTOM_TRAIN = 1;            }
    elsif (/^-T$/ ||
	   /^--test$/) { $TEST_PROP = shift @ARGV; 
			 $CUSTOM_TEST = 1;             }
    elsif (/^-i$/)     { $INT_PROP = shift @ARGV;      }
    elsif (/^-E$/)     { $ENG_SUFFIX = shift @ARGV;    }
    elsif (/^-F$/)     { $FOR_SUFFIX = shift @ARGV;    }
    else { warn "Ignoring unknown parameter $_\n";     }
  }

  # check for valid parameters
  undef $MAX_LENGTH if ($MAX_LENGTH && $MAX_LENGTH <= 0);
  undef $MIN_FREQ if ($MIN_FREQ && $MIN_FREQ <= 1);
  die "$TEST_PROP is invalid as a test set proportion.\n"
    if ($TEST_PROP == 0 && $TEST_PROP ne "0");
  die "$INT_PROP is invalid as a interleaving proportion.\n"
    if ($INT_PROP == 0 && $INT_PROP ne "0");
  if ($TEST_ONLY && $TRAIN_ONLY) {
    undef $TEST_ONLY;
    undef $TRAIN_ONLY;
    print "\nNote: --test --train is identical to the default setting.\n";
  }
  die "Unknown word ID must be smaller than first word ID.\n"
    if ($FIRST_WORD_ID <= $UNKNOWN_ID);
  if ($NOSPLIT) {
    $BASE_RATIO = $TEST_PROP = 0;
    $TRAIN_PROP = $TRAIN_ONLY = 1;
    undef $TEST_ONLY;
  }
  
  # now build file list
  
  $ARGV[0] = "." unless $ARGV[0];
  
  while ($ARGV[0]) {
    $_ = shift @ARGV;
    $_ .= "/" unless (!-d || m|/\Z|); # ensure trailing '\'
    print "\nExamining directory $_\n";
    $IN_DIR = $_ unless $IN_DIR;
    if (0 && !-d) {
      warn "$ARGV[0] doesn't appear to be a " . 
	"directory. Ignoring...\n";
      next;
    }
    push(@FOR_FILES, glob("$_*$FOR_SUFFIX"));
    push(@ENG_FILES, glob("$_*$ENG_SUFFIX"));
  }
  
  check_files();

  if ($OUTFILE =~ m|(.*/)([^/]*)\Z|) {
    $OUT_DIR = $1;
    $OUTFILE = $2;
  }
  if (!$OUTFILE) {
    $OUTFILE = $ENG_FILES[0];
    $OUTFILE =~ s/.*\///;
    $OUTFILE =~ s/$ENG_SUFFIX\Z//;
  }
  $OUT_DIR = $IN_DIR unless ($OUT_DIR || $OUTFILE =~ m|/|);
  
  # print our operating parameters
  
  print "\nFile list:\n----------\n";
  for ($i = 0; $i < scalar(@ENG_FILES); $i++) {
    print file_name($ENG_FILES[$i])."\t".
      file_name($FOR_FILES[$i])."\n";
  }

  print "\n";
  print "Using options:\n";
  print_options();
  
  # now init globals for faster access
  $SNT = ($FORMAT && ($FORMAT eq ".snt"));
  $BST = ($FORMAT && ($FORMAT eq ".bst"));
  $FORMAT || ($INTEGERIZE = 0);
  $BST && !$INTEGERIZE &&
    die "sorry, binary output only available with integerization.\n";
}

sub process {
  my ($e, $f, $out) = @_;
  my (@ew, @fw);

  return undef if (!$$e || !$$f);

  chomp($$e);
  chomp($$f);
  @ew = split(/\s+/, $$e);
  @fw = split(/\s+/, $$f);
  shift @ew unless length($ew[0]);
  shift @fw unless length($fw[0]);
  return undef if ($MAX_LENGTH && 
		   ($#ew >= $MAX_LENGTH || 
		    $#fw >= $MAX_LENGTH));
  return \ ($$e . "\n" . $$f . "\n") 
    unless ($INTEGERIZE || $MIN_FREQ);
  return \ (${integerize(\@ew, $ENG_TABLE)} . "\n" . 
    ${integerize(\@fw, $FOR_TABLE)} . "\n");
}

sub integerize {
  my ($words, $table) = @_;
  my ($w, $f, $out);
  if ($INTEGERIZE) {
    for $w (@$words) {
      if (defined($f = $$table{$w})) {
	$w = $f;
      } else {
	$w = $UNKNOWN_ID;
      }
    }
  } else {
    for $w (@$words) {
      $w = $UNKNOWN_WORD unless (defined($$table{$w}))
    }
  }
  return \join(' ', @$words);
}

# load count files using clsub, looking in many different places.
# checks in the input directory, output directory, and current directory
# before automatically creating new ones.
sub count_load {
  my ($elen, $flen);
  
  print "\nchecking for count files in $IN_DIR\n";
  ($elen, $ENG_TABLE) = clsub(cnt_stem($IN_DIR, $ENG_SUFFIX),
			      @ENG_FILES);
  ($flen, $FOR_TABLE) = clsub(cnt_stem($IN_DIR, $FOR_SUFFIX),
			      @FOR_FILES);
  unless ($elen && $flen) {
    print "none found. checking for count files in $OUT_DIR\n";
    ($elen, $ENG_TABLE) = clsub(cnt_stem($OUT_DIR, $ENG_SUFFIX),
				@ENG_FILES);
    ($flen, $FOR_TABLE) = clsub(cnt_stem($OUT_DIR, $FOR_SUFFIX),
				@FOR_FILES);
    unless ($elen && $flen) {
      print "none found. checking for count files in ./\n";
      ($elen, $ENG_TABLE) = clsub(cnt_stem("./", $ENG_SUFFIX),
				  @ENG_FILES);
      ($flen, $FOR_TABLE) = clsub(cnt_stem("./", $FOR_SUFFIX),
				  @FOR_FILES);
      unless ($elen && $flen) {
	print "no count files found. ";
	count();
	($elen, $ENG_TABLE) = clsub(cnt_stem($IN_DIR, $ENG_SUFFIX),
				    @ENG_FILES);
	($flen, $FOR_TABLE) = clsub(cnt_stem($IN_DIR, $FOR_SUFFIX),
				    @FOR_FILES);
	unless ($elen && $flen) {
	  ($elen, $ENG_TABLE) = clsub(cnt_stem($OUT_DIR, $ENG_SUFFIX),
				      @ENG_FILES);
	  ($flen, $FOR_TABLE) = clsub(cnt_stem($OUT_DIR, $FOR_SUFFIX),
				      @FOR_FILES);
	}
      }
    }
  }
  warn "WARNING: DIFFERENT NUMBERS OF SENTENCES IN PARALLEL CORPORA!\n" . 
    "    This is typically a Very Bad Sign.\n" 
      if ($elen != $flen);
  
  $NUM_SENT = (($elen <= $flen) ? $elen : $flen);
}

# takes care of reading in one pair of .cnt and .vcb files
# first argument is the stem of the files, the rest is a list
# of all files of this language in the input directory
sub clsub {
  my ($line, @in, $a, $b, $i, $n, %table);
  my $stem = shift;
  # first read the count file and get the sentence count
  open(IN, "<$stem$COUNT_SUFFIX") || return undef;
  print "reading $stem$COUNT_SUFFIX\n";
  $line = <IN>;
  $line =~ /(\d+)\s/;
  $n = $1;
  # now discard all until first blank line
  $line = <IN> while (length($line)>1);
  
  while ($line = lc(<IN>)) {
    chomp($line);
    last unless $line;
    push(@in, $line);
  }
  
  close(IN);
  # now check to make sure file lists match
 MARK1:
  for $a (@_) {
    for $b (@in) {
      next MARK1 if ($b eq lc(file_name($a)));
    }
    warn "WARNING: the file $a does not appear " . 
      "in the count file $stem$COUNT_SUFFIX\n";
  }
 MARK2:
  for $b (@in) {
    for $a (@_) {
      next MARK2 if ($b eq lc(file_name($a)));
    }
    warn "WARNING: the file $b in the count file $stem$COUNT_SUFFIX\n" .
      "do not appear in the input directory.\n";
  }
  
  # if we're not integerizing, no need to read vocab file
  return $n unless ($INTEGERIZE || $MIN_FREQ);
  
  # otherwise, read vcb file
  # format is: word_id word word_frequency on each line
  print "reading $stem$VOCAB_SUFFIX\n";
  open(IN, "<$stem$VOCAB_SUFFIX") || return undef;
  my ($id, $w, $freq, $over);
  while (<IN>) {
    chomp;
    ($id, $w, $freq, $over) = split;
    unless (length($id) && length($w) && 
	    length($freq) && !defined($over)) {
      warn "Bad line $. in $stem$VOCAB_SUFFIX: $_\n";
      next;
    }
    $table{$w} = $id unless ($MIN_FREQ && $freq < $MIN_FREQ);
  }
  close(IN);
  # return a reference to the table; it's faster.
  return ($n, \%table);
}

# runs the count procedure; depends on csub
sub count {
  my(@ef, @ff, %freq, @words, $ewc, $esc, $fwc, $fsc);

  print "\nRunning count on English corpus files:\n";
  ($ewc, $esc) = csub($ENG_SUFFIX, @ENG_FILES);# @ef);
  print "\nRunning count on foreign corpus files:\n";
  ($fwc, $fsc) = csub($FOR_SUFFIX, @FOR_FILES);# @ff);
  
  print "\nEnglish corpus: $ewc words, $esc sentences.\n";
  print "Foreign corpus: $fwc words, $fsc sentences.\n";
  warn "WARNING: DIFFERENT NUMBERS OF SENTENCES IN CORPORA!\n" . 
    "    This is typically a Very Bad Sign.\n" 
      if ($COUNT_ONLY && ($esc != $fsc));
  print "count done.\n";
}

sub csub {
  my ($i, $w, $p, @words, $wc, $sc, %freq);
  my $suffix = shift;
  my @files = @_;
  for $p (@files) {
    open(A, "<$p") || die "Unable to read file $p\n";
    print "reading $p\n";
    while (<A>) {
      chomp;
      @words = split;
      for $w (@words) {
	# if($w) doesn't work because of token '0'
	if (length($w)) {
	  $freq{$w}++;
	  $wc++;
	}
      }
      $sc++;
    }
    close(A);
  }

  print "sorting words by frequency\n";
  @words = sort {
    $freq{$b} <=> $freq{$a}
  } keys %freq;
  
  $p = cnt_stem($IN_DIR,$suffix) . $COUNT_SUFFIX;
  print "creating count file $p...";
  unless (open(OUT, ">$p")) {
    $p = cnt_stem($OUT_DIR,$suffix) . $COUNT_SUFFIX;
    print "failed\ncreating count file $p...";
    open(OUT, ">$p") || die "failed\nwhittle aborting.\n";
  }
  print OUT $sc . " sentences\n";
  print OUT $wc . " words\n";
  print OUT scalar(@words) . " unique\n";
  print OUT "\n";
  for $p (@files) {
    $p =~ /.*\/([^\/]+)\Z/;
    print OUT "$1\n";
  }
  close(OUT);
  print "done.\n";
  
  $p =~ s/$COUNT_SUFFIX\Z/$VOCAB_SUFFIX/;
  print "creating vocab file $p...";
     
  open(OUT, ">$p") || die "failed.\nwhittle aborting.\n";
  print OUT "$UNKNOWN_ID\t$UNKNOWN_WORD\t0\n";
  $i = $FIRST_WORD_ID;
  for $w (@words) {
    print OUT $i++ . "\t$w\t$freq{$w}\n";
  }
  close(OUT);
  print "done.\n";

  return ($wc, $sc);
}

# ensures that every english file has a foreign partner,
# and vice versa
# puts result in 2D global array @FILES
sub check_files() {
  my ($t, $e, $f, $te, $tf, $i, $j, $k);
  $i = $j = $k = 0;
  print "Verifying file list...\n";
  die "No English ($ENG_SUFFIX) files found in directory $IN_DIR\n"
    if ($#ENG_FILES < 0);
  die "No foreign ($FOR_SUFFIX) files found in directory $IN_DIR\n"
    if ($#FOR_FILES < 0);
  
  while ($ENG_FILES[$i] && $FOR_FILES[$j]) {
    $f = $FOR_FILES[$j];
    $t = $e = $ENG_FILES[$i];
    
    $t =~ s/$ENG_SUFFIX\z/$FOR_SUFFIX/;
    # THIS IS A HACK TO NOT USE THE EMPTY FILES
    #if ($f =~ /136/ || $f =~ /159/) {
    #  splice(@ENG_FILES, $i, 1);
    #  splice(@FOR_FILES, $j, 1);
    #  next;
    #} # END HACK
    if (0) { #$f ne $t) {
    TEST:
      for $te (@ENG_FILES) {
	$te =~ s/$ENG_SUFFIX\z/$FOR_SUFFIX/;
	for $tf (@FOR_FILES) {
	  last TEST if ($te eq $tf);
	}
      }
      $te =~ s/$FOR_SUFFIX\z/$ENG_SUFFIX/;
      if ($e ne $te) {
	while ($ENG_FILES[$i] ne $te) {
	  $t = splice(@ENG_FILES, $i, 1);
	  print "Discarding English file $t with no parallel file.\n";
	}
	$e = $te;
      }
      if ($f ne $tf) {
	while ($FOR_FILES[$j] ne $tf) {
	  $t = splice(@FOR_FILES, $j, 1);
	  print "Discarding foreign file $t with no parallel file.\n";
	}
	$f = $tf;
      }
    }
    $i++;
    $j++;
  }

  # now remove extra files from end of arrays
  print "discarding English file " . pop(@ENG_FILES) . 
    " with no parallel file.\n" while ($ENG_FILES[$i]);
  
  print "discarding foreign file " . pop(@FOR_FILES) . 
    " with no parallel file.\n" while ($FOR_FILES[$j]);
}



sub print_options {
  print "\tOutput format:            ".
    ($FORMAT ? $FORMAT : "parallel")."\n";
  print "\tIntegerizing:             " . 
    ($INTEGERIZE ? "on" : "off") . "\n";
  print "\tMax sentence length:      " . 
    ($MAX_LENGTH ? $MAX_LENGTH : "none") . "\n";
  print "\tMin word frequency:       " .
    ($MIN_FREQ ? $MIN_FREQ : "none") . "\n";
  print "\tUnknown word token:       $UNKNOWN_WORD\n";
  print "\tUnknown word id:          $UNKNOWN_ID\n";
  print "\tFirst word id:            $FIRST_WORD_ID\n";
  print "\tOutput corpus ".($CORPUS_PROP > 1 ? "size: " : "ratio:").
    "      $CORPUS_PROP\n";
  print "\tBaseline test set ratio:  ".
    ($NOSPLIT ? $NA : $BASE_RATIO)."\n";
  print "\tTraining set ".($TRAIN_PROP > 1 ? "size: " : "ratio:").
    "       ".($NOSPLIT ? $NA : $TRAIN_PROP)."\n";
  print "\tTest set ".($TEST_PROP > 1 ? "size: " : "ratio:").
    "           ".($NOSPLIT ? $NA : $TEST_PROP)."\n";
  print "\tInterleaving ratio:       ".
    ($NOSPLIT ? $NA : $INT_PROP)."\n";
  print "\tEnglish file suffix:      $ENG_SUFFIX\n";
  print "\tForeign file suffix:      $FOR_SUFFIX\n";
  print "\tOutput stem:              $OUTFILE\n" if ($OUTFILE);
  print "\tOutput directory:         $OUT_DIR\n" if ($OUT_DIR);
}

sub print_corp_info {
  print "$NUM_SENT total sentences in input files\n";
  print "Using ".($train_ratio ? "1 of every $train_ratio" : "all").
    " training sentences.\n";
  print "Using ".($test_ratio ? "1 of every $test_ratio" : "all").
    " test sentences.\n";

  print "".int($corp_size)." total sentences in output corpus\n";
  print "".($TEST_ONLY ? 0 : int($train_size)).
    " training sentences\n" unless $NOSPLIT;
  print "".($TRAIN_ONLY ? 0 : int($test_size)).
    " test sentences\n" unless $NOSPLIT;
  print "Interleaving ratio 1 test sentence per $int_ratio training\n\n";
}

sub write_corp_stats {
  my ($train, $test) = @_;
  open(B, ">$OUT_DIR$OUTFILE.stats") || 
    die "couldn't write to $OUT_DIR$OUTFILE.stats\n";
  print "writing corpus stats to $OUT_DIR$OUTFILE.stats\n";
  print B "$NUM_SENT total sentences in input files\n";
  print B ($train+$test) . " total sentences in output corpus\n";
  print B ($NOSPLIT ? $NA : $BASE_RATIO) . " baseline ratio\n";
  print B ($NOSPLIT ? $NA : $train) . " training sentences\n";
  print B ($NOSPLIT ? $NA : $test) . " test sentences\n";
  print B ($NOSPLIT ? $NA : ($train_ratio ? $train_ratio : 1)).
    " training sentence selection ratio\n";
  print B ($NOSPLIT ? $NA : ($test_ratio ? $test_ratio : 1)).
    " test sentence selection ratio\n";
  print B ($NOSPLIT ? $NA : $int_ratio).
    " test sentence interleaving ratio\n";
  print B ($MIN_FREQ ? $MIN_FREQ : 0).
    " minimum word frequency\n";
  print B ($MAX_LENGTH ? $MAX_LENGTH : 0).
    " maximum sentence length\n";
  close(B);
}

sub write_snt {
  my ($which, $counts) = @_; # counts is hashref
  my $o = out_name($which);
  my $t = temp_name($which);
  my $n = 0;
  my ($pair);
  
  open(IN, $t) || die "couldn't open temp file $t\n";
  #seek(IN, 0, 0);
  open(OUT, ">$o") # clobber if necessary
    || die "couldn't write to $o\n";
  print "writing $o\n";

  while ($pair = <IN> . <IN>) {
    print OUT delete($$counts{$pair}) . "\n$pair";
    print "wrote $n pairs\n" if (++$n % 100000 == 0);
  }
  close OUT;
  close IN;
  unlink($t)
    || warn "WARNING: could not delete temp file $t\n";
}

sub out_name {
  my $stem = shift;
  $stem = '-'.$stem if $stem;
  return ($OUT_DIR.$OUTFILE.$stem.$FORMAT);
}
sub cnt_stem {
  return shift().substr(shift(),1);
}
sub temp_name {
  my $stem = shift;
  $stem = '-'.$stem if $stem;
  return ($OUT_DIR.$TEMP_PREFIX.$stem.$TEMP_SUFFIX);
}
sub file_name {
  shift =~ m|.*/([^/]+)\Z|;
  return $1;
}
